#include <windows.h>
#include "MQPlugin.h"
#include <utility>
#include <memory>
#include <string>
#include <vector>

class CMQCoordinate
{
public:
	CMQCoordinate() : index_(-1) {cood_.u =cood_.v = 0.0f;}
	CMQCoordinate(CMQCoordinate &&cood) : index_(cood.index_), cood_(cood.cood_) {}
	CMQCoordinate &operator=(CMQCoordinate &&cood) {index_ = cood.index_; cood_ = cood.cood_; return *this;}

	CMQCoordinate(int index, const MQCoordinate &cood) : index_(index), cood_(cood) {}
	bool operator==(const CMQCoordinate &other) const {return index_ == other.index_;}
	static bool uvcomp(const CMQCoordinate &left, const CMQCoordinate &right) {
		return pow2(left.cood_.u - right.cood_.u) + pow2(left.cood_.v - right.cood_.v) < pow2(uvdelta);
	}
	int getIndex() const {return index_;}
	MQCoordinate getCoordinate() const {return cood_;}
private:
	int index_;
	MQCoordinate cood_;

	static const float uvdelta; // UVlꔻp
	static float pow2(float v) {return v * v;}
};

class CMQObject
{
friend class CMQDocument;
friend class Vertex;
friend class Face;
public:
	CMQObject(MQObject obj = nullptr) : obj_(obj) {}
	CMQObject(CMQObject &&obj) : obj_(obj.obj_) {}
	CMQObject &operator=(CMQObject &&obj) {obj_ = obj.obj_; return *this;}

	operator bool() const {return obj_ != nullptr;}
	CMQObject clone();
	std::string getName();
	void setName(const std::string &name);
	bool getSelected() {return obj_->GetSelected() == TRUE;}
	bool getLocking() {return obj_->GetLocking() == TRUE;}

	int getVertexCount() {return obj_->GetVertexCount();}
	int getVertexRefCount(int vindex) {return obj_->GetVertexRefCount(vindex);}
	MQPoint getVertex(int vindex) {return obj_->GetVertex(vindex);}
	UINT getVertexUniqueID(int vindex) {return obj_->GetVertexUniqueID(vindex);}
	int getVertexIndexFromUniqueID(UINT uid) {return obj_->GetVertexIndexFromUniqueID(uid);}
	void setVertex(int vindex, MQPoint p) {obj_->SetVertex(vindex, p);}

	int getFaceCount() {return obj_->GetFaceCount();}
	int getFacePointCount(int findex) {return obj_->GetFacePointCount(findex);}
	bool getFaceVisible(int findex) {return obj_->GetFaceVisible(findex) == TRUE;}

	std::vector<int> getVertexRelatedFaces(int vindex);
	std::vector<int> getFacePointArray(int findex);
	std::vector<CMQCoordinate> getFaceCoordinateArray(int findex);
	void setFaceCoordinateArray(int findex, int vertex, const CMQCoordinate &co);

private:
	MQObject obj_;
};

class CMQDocument
{
friend class CMQObject;
friend class Vertex;
friend class Face;
public:
	CMQDocument(MQDocument doc) : doc_(doc) {}
	CMQDocument(CMQDocument &&doc) : doc_(doc.doc_) {}
	CMQDocument &operator=(CMQDocument &&doc) {doc_ = doc.doc_; return *this;}

	CMQObject getCurrentObject();
	CMQObject &insertObject(CMQObject &newobj, const CMQObject &before);
	void deleteObject(int i) {doc_->DeleteObject(i);}
	void deleteObject(CMQObject obj) {int i = getObjectIndex(obj); if(i == -1) obj.obj_->DeleteThis(); else doc_->DeleteObject(i);}
	std::string getUnusedName(const std::string &name);
	int getObjectCount() {return doc_->GetObjectCount();}
	CMQObject getObject(int i) {return CMQObject(doc_->GetObject(i));}
	int getObjectIndex(CMQObject obj) {return doc_->GetObjectIndex(obj.obj_);}
	CMQObject getObjectbyName(const std::string &name);

	bool isSelectVertex(CMQObject obj, int vindex) {return doc_->IsSelectVertex(doc_->GetObjectIndex(obj.obj_), vindex) == TRUE;}
	bool addSelectVertex(CMQObject obj, int vindex) {return doc_->AddSelectVertex(doc_->GetObjectIndex(obj.obj_), vindex) == TRUE;}
	bool deleteSelectVertex(CMQObject obj, int vindex) {return doc_->DeleteSelectVertex(doc_->GetObjectIndex(obj.obj_), vindex) == TRUE;}

	bool isSelectFace(CMQObject obj, int findex) {return doc_->IsSelectFace(doc_->GetObjectIndex(obj.obj_), findex) == TRUE;}
	bool addSelectFace(CMQObject obj, int findex) {return doc_->AddSelectFace(doc_->GetObjectIndex(obj.obj_), findex) == TRUE;}
	bool deleteSelectFace(CMQObject obj, int findex) {return doc_->DeleteSelectFace(doc_->GetObjectIndex(obj.obj_), findex) == TRUE;}

	void clearSelect(DWORD flag) {doc_->ClearSelect(flag);}

private:
	MQDocument doc_;
};
